/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Boucle de jeu : dessin et gestion des donnes.
 *****************************************************************************/

#include "game_loop.h"
//-----------------------------------------------------------------------------
#include "../graphic/fps.h"
#include "../graphic/graphism.h"
#include "../include/constant.h" // FOND_X, FOND_Y
#include "../interface/game_msg.h"
#include "../interface/mouse.h"
#include "../interface/keyboard.h"
#include "../interface/cursor.h"
#include "../interface/interface.h"
#include "../object/objects_list.h"
#include "../object/bonus_box.h"
#include "../map/camera.h"
#include "../map/wind.h"
#include "../map/map.h"
#include "../map/maps_list.h"
#include "../sound/jukebox.h"
#include "../team/macro.h"
#include "../tool/i18n.h"
#include "../weapon/weapons_list.h"
#include "game.h"
#include "time.h"
#include "debug.h"
#include "game_mode.h"
#include "config.h"
#include <sstream>
using namespace Wormux;
//-----------------------------------------------------------------------------

#ifdef DEBUG

// Debogue la fin du tour ?
//#define DBG_FIN_TOUR

// Afiche les changements d'tat ?
//#define DEBUG_CHG_ETAT

#endif

#define COUT_DEBUG cout << "[BoucleJeu] "

const unsigned int MAX_FPS = 50;
const unsigned int SLEEP_LIMIT_FPS = 1000/MAX_FPS;

//-----------------------------------------------------------------------------
BoucleJeu boucle_jeu;
//-----------------------------------------------------------------------------

void InitJeu ()
{
  // Message pour indiquer le chargement d'une partie
  std::cout << std::endl;
  std::cout << "[ " << _("Starting a new game") << " ]" << std::endl;
  jeu.MsgChargement();

  if (!jeu.initialise)
  {
    std::cout << "o " << _("Initialisation") << std::endl;
    monde.Init();
    eau.Init();
    interface.Init();
    curseur_ver.Init();
    lst_objets.Init();
    //#ifdef WEAPON_CLEAN
    // RE-loading game_mode ...
    if (!mode_jeu.Charge(config.data_dir+"game_mode/classic.xml"))
      std::cout << "Error loading game mode !" << std::endl;
    //#endif

    jeu.initialise = true;
  }

  // Placement des vers
  std::cout << "o " << _("Find a random position for worms") << std::endl;
  monde.Reset();
  eau.Reset();
  eau.Active(lst_terrain.TerrainActif().use_water);
  teams_list.Reset();

  // Remise  zro
  std::cout << "o " << _("Initialise data") << std::endl;
  curseur_ver.Reset();
  temps.Reset();
  vent.Reset();
  souris.Reset();
  image_par_seconde.Reset();
  interface.Reset();
  messages_jeu.Reset();
  lst_objets.Reset();

  // Cration des quipes
  if (teams_list.list.size() < 2)
    Erreur(_("You need at least two teams to play: "
	     "change this in 'Options menu' !"));
  assert (teams_list.list.size() <= mode_jeu.nbr_equipe_max);

  // Initialisation de l'nergie d'nergie de chaque quipe
  teams_list.InitEnergy();

  // Charge l'ensemble des sons des equipes
  jukebox.Load("default");
  POUR_CHAQUE_EQUIPE(equipe) 
    if ( (**equipe).LitProfileSon() != "default" )
      jukebox.Load((**equipe).LitProfileSon()) ;  

  // Dbut le premier tour !
  jeu.fin_partie = false;
  boucle_jeu.ChangeEtat (jeuVER_JOUE, true);

  // Joue une musique d'ambiance
  if (jukebox.UseMusic()) jukebox.Play ("ambiance/grenouilles", true);
}

//-----------------------------------------------------------------------------

BoucleJeu::BoucleJeu()
{
  etat = jeuVER_JOUE;
}

//-----------------------------------------------------------------------------

void BoucleJeu::Refresh()
{
  // Refresh du camera
  camera.Refresh();

  // Mise  jeu des entres (clavier / souris)
#ifdef ECONOMISE_CPU
  CL_System::keep_alive(ECONOMISE_CPU);
#else
  CL_System::keep_alive();
#endif

  // Compteur du nombre d'images par seconde
  image_par_seconde.Refresh();

  //--- D'abord ce qui pourrait modifier les donnes d'un ver ---

  if (!temps.EstPause())
  {
    // Refresh de la souris et du clavier
    if (jeu.autorise_action &&
	((etat == jeuVER_JOUE) 
	 || (etat == jeuVER_A_JOUE)
	 || ((etat == jeuPAUSE_FIN_TOUR))))
    {
      souris.Refresh();
      clavier.Refresh();
    }

    // Refresh des vers
    POUR_TOUS_VERS(equipe,ver) ver -> Refresh();

    // Refresh des vers
    POUR_CHAQUE_EQUIPE(team) (**team).Update();
    
    //--- Ensuite, actualise le reste du jeu ---

    // Refresh des deplacements et actions des vers
    gst_armes.Refresh();

    // Dessin tous les objets
    lst_objets.Refresh();

    // Curseur
    curseur_ver.Refresh();
  }
  
  teams_list.RefreshEnergy();

  // Refresh du monde
  eau.Refresh();

#ifdef DEBUG
  // Draw les messages de debug
  debug.Refresh();
#endif
}

//-----------------------------------------------------------------------------

void BoucleJeu::Run()
{
  do
  {
	unsigned int start = CL_System::get_time();
	  
    jeu.fin_partie = false;
    BeginLoop();
    Refresh();
    EndLoop();
    CallDraw ();

	unsigned int delay = CL_System::get_time()-start;
	if (delay < SLEEP_LIMIT_FPS) CL_System::sleep(SLEEP_LIMIT_FPS - delay);
  } while (!jeu.fin_partie);
}

//-----------------------------------------------------------------------------

void BoucleJeu::CallDraw()
{
  Draw(jeu.force_dessin);
  CL_Display::flip (false);
  jeu.force_dessin = false;
}

//-----------------------------------------------------------------------------

void BoucleJeu::Draw (bool force)
{
  std::ostringstream txt;

  CL_Display::push_cliprect( CL_Rect(FOND_X, FOND_Y, 
  					  FOND_X+camera.GetWidth(), 
					  FOND_Y+camera.GetHeight()));

  monde.DrawSky();

  CL_Display::push_translate((int)FOND_X-camera.GetX(), 
			     (int)FOND_Y-camera.GetY());

  vent.DrawParticles();

  // Redessine uniquement le dcor si on l'a boug
  force = true;
  monde.Draw();

  // Dessine les vers
  POUR_TOUS_VERS(equipe,ver) if (&(*ver) != &ActiveCharacter()) ver -> Draw();
  ActiveCharacter().Draw();

  if (!ActiveCharacter().IsDead()) 
  {
    if ((etat == jeuVER_JOUE) 
	|| (etat == jeuVER_A_JOUE) 
	|| (etat == jeuANIMATION))
      ActiveTeam().AccesArme().Draw();
  }

  lst_objets.Draw();
  ActiveTeam().viseur.Draw();
  curseur_ver.Draw();
  eau.Draw();

  // Draw team informations
  POUR_CHAQUE_EQUIPE(team) (**team).Draw();

  // Supprime le decalage
  CL_Display::pop_modelview();

  // Affiche les messages
  messages_jeu.Affiche();

  // Affiche le nom de l'auteur du terrain
  monde.DrawAuthorName();

  // Affiche le nombre d'images/seconde
  image_par_seconde.Draw();

  // Affiche l'horloge du jeu
  temps.Draw();
  
  // Supprime le clipping
  CL_Display::pop_cliprect();

#ifdef DEBUG
  // Draw les messages de debug
  debug.Draw();
#endif

  // Draw l'interface
  interface.Draw (force); 

  // Dessine la barre de force du vent
  vent.Draw();

  // Ca fait une image de plus pour le compteur ;-)
  image_par_seconde.AjouteUneImage();
}

//-----------------------------------------------------------------------------

// Preparation avant un nouveau cycle de boucle
void BoucleJeu::BeginLoop()
{
  if (temps.EstPause()) return;

  switch (etat)
  {
  // Un ver joue
  case jeuVER_JOUE:
    // Actualise le compteur du temps restant d'un tour
    if (1000 < temps.Lit() - pause_seconde)
    {
      pause_seconde = temps.Lit();
      if (tps_tour == 1) {
	jukebox.Play("end_turn");
	jeu.autorise_action = false;
	boucle_jeu.ChangeEtat (jeuPAUSE_FIN_TOUR);
      } else {
	tps_tour--;
	interface.chrono = tps_tour;
      }
    }
    break;

  // Fin d'un tour
  case jeuPAUSE_FIN_TOUR:
    // Actualise le compteur du temps restant d'un tour
    if (1000 < temps.Lit() - pause_seconde)
    {
      pause_seconde = temps.Lit();
      if (tps_fin_tour == 1) {
	if (jeu.EstFinPartie()) 
	  jeu.fin_partie = true;
	else
	  ChangeEtat (jeuANIM_FIN_TOUR);
      } else {
	tps_fin_tour--;
	interface.chrono = tps_fin_tour;
      }
    }
    break;

  default: break;
  }

  // Refresh des messages
  messages_jeu.Actualise();
}

//-----------------------------------------------------------------------------

void BoucleJeu::DebutTour()
{
  tps_tour = mode_jeu.temps_tour;
  pause_seconde = temps.Lit();
  vent.ChangeVal();
}

//-----------------------------------------------------------------------------

void BoucleJeu::FinTour()
{
  tps_fin_tour = mode_jeu.temps_fin_tour;
  pause_seconde = temps.Lit();
}

//-----------------------------------------------------------------------------

void BoucleJeu::ChangeEtat (etat_jeu nv_etat, bool debut_jeu)
{
  // Si on est dj dans le bon tat, ne fait rien
  if ((etat == nv_etat) && !debut_jeu) return;

  // Change d'tat
  etat = nv_etat;

  // Action selon le nouvel tat
  switch (nv_etat)
  {
  // Dbut d'un tour
  case jeuVER_JOUE:
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "Etat un jeu joue." << endl;
#endif
    // Init. le compteur
    interface.chrono = mode_jeu.temps_tour;
    boucle_jeu.DebutTour();
    jeu.choix_ver_fait = false;
    jeu.autorise_action = true;
    pause_fin_tour = false;

    // Prpare un tour pour un ver
    POUR_TOUS_VERS_VIVANTS(equipe,ver) ver -> PrepareTour();

    // Changement d'quipe
    assert (!jeu.EstFinPartie());
    teams_list.NextTeam (debut_jeu);
    assert (!ActiveCharacter().IsDead());

    // Si le ver n'est pas visible, se recentre sur lui
    if (!camera.EstVisible(ActiveCharacter()))  
      camera.ChangeObjSuivi (&ActiveCharacter(), true, true);
    break;

  // Un ver a jou son arme, mais peut encore se dplacer
  case jeuVER_A_JOUE:
    interface.chrono = -1;
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "Etat un ver a joue" << endl;
#endif
    break;

  // Animation d'une arme
  case jeuANIMATION:
    interface.chrono = -1;
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "Etat animation" << endl;
#endif
    break;

  // Fin du tour : petite pause
  case jeuPAUSE_FIN_TOUR:
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "Etat pause de fin de tour" << endl;
#endif
    if (EndLoop()) break;
    pause_fin_tour = true;
    ActiveTeam().AccesArme().Desactive();
    boucle_jeu.FinTour();
    interface.chrono = mode_jeu.temps_fin_tour;
    break;

  // Animation en fin du tour
  case jeuANIM_FIN_TOUR:
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "Etat animation de fin de tour" << endl;
#endif
    interface.chrono = -1;
    caisse.FaitApparaitre();
    EndLoop();
    break;
  }
}

//-----------------------------------------------------------------------------

PhysicalObj* ChercheObjBouge()
{
  if (!ActiveCharacter().IsReady()) return &ActiveCharacter();

  POUR_TOUS_VERS(equipe,ver)
  {
    if (!ver -> IsReady())
    {
#ifdef DBG_FIN_TOUR
      cout << (*ver).nom << " n'est pas prt" << endl;
#endif
      return &(*ver);
    }
  }

  POUR_CHAQUE_OBJET(objet)
  {
    if (!objet -> ptr -> IsReady())
    {
#ifdef DBG_FIN_TOUR
      cout << objet -> ptr -> nom << " n'est pas prt" << endl;
#endif
      return objet -> ptr;
    }
  }

  return NULL;
}

//-----------------------------------------------------------------------------

bool BoucleJeu::EndLoop()
{
  bool bouge_encore = false;

  if (etat == jeuVER_JOUE) return false;

  if (ActiveTeam().LitArme().IsActive()) bouge_encore = true;

  if (!bouge_encore)
  {
    PhysicalObj *obj = ChercheObjBouge();
    if (obj != NULL)
    {
      camera.ChangeObjSuivi (obj, true, true);
      bouge_encore = true;
    }
  }

  if (bouge_encore) {
    if ((etat == jeuPAUSE_FIN_TOUR) && !pause_fin_tour) 
      { ChangeEtat (jeuANIMATION); return true; }
  } else {
    if (etat == jeuANIMATION) { ChangeEtat (jeuPAUSE_FIN_TOUR); return true; }
    if (etat == jeuANIM_FIN_TOUR) { ChangeEtat (jeuVER_JOUE); return true; }
    if (etat == jeuVER_A_JOUE) { ChangeEtat (jeuPAUSE_FIN_TOUR); return true; }
  }
  return false;
}

//-----------------------------------------------------------------------------
