/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Abstract class used for physical object (object with a size, mass,
 * etc.). This object can have differents state : ready, is moving, or ghost
 * (is outside of the world).
 *
 * You can : make the object move (with collision test), change state, etc.
 * If the object go outside of the world, it become a ghost.
 *****************************************************************************/

#include "../object/physical_obj.h"
//-----------------------------------------------------------------------------
#include <stdlib.h>
#include "../map/map.h"
#include "../game/config.h"
#include "../tool/math_tools.h"
#include "../tool/geometry_tools.h"
#include "../game/time.h"
#include "../map/water.h"
#include "../sound/jukebox.h"
#include "../map/wind.h"
using namespace Wormux;
//-----------------------------------------------------------------------------

#ifdef DEBUG

//#define DEBUG_CHG_ETAT
//#define DEBUG_MSG_PHYSIQUE

#define COUT_DEBUG0 std::cout << "[Objet " << m_name << "]"
#define COUT_DEBUG COUT_DEBUG0 " "
#define COUT_PHYSIC COUT_DEBUG0 << "[Physique] "

#endif
//-----------------------------------------------------------------------------

const int Y_OBJET_MIN = -10000;

//-----------------------------------------------------------------------------

PhysicalObj::PhysicalObj (const std::string &name, double mass)
  : m_name(name),
    m_width(0),
    m_height(0)
{
  m_type = objCLASSIC;
//  m_moving = false;
  m_pos.x = m_pos.y = 0;
  m_strength.x = m_strength.y = 0;
  m_speed.x = m_speed.y = 0;
  
  m_allow_negative_y = false;
  m_alive = ALIVE;

  m_rebounding = false;
  m_rebound_factor = 0.01;
  m_rebound_sound = "";
  
  m_wind_factor = 1.0;

  m_test_left = 0;
  m_test_right = 0;
  m_test_top = 0;
  m_test_bottom = 0;

  m_ready = true;

  SetMass(mass);
  m_last_move = Wormux::temps.Lit();
}

//-----------------------------------------------------------------------------

PhysicalObj::~PhysicalObj ()
{}

//-----------------------------------------------------------------------------

void PhysicalObj::SignalGhostState(bool)  {}
void PhysicalObj::SignalDeath() {}
void PhysicalObj::SignalDrowning() {}
void PhysicalObj::SignalFallEnding() {}

//-----------------------------------------------------------------------------

const CL_Point PhysicalObj::GetPos() const 
{ 
  return CL_Point(GetX(), GetY()); 
}

double PhysicalObj::GetXf() const 
{ 
  assert (!IsGhost());
  return m_pos.x; 
}
double PhysicalObj::GetYf() const 
{ 
  assert (!IsGhost());
  return m_pos.y; 
}
int PhysicalObj::GetX() const { return Arrondit(GetXf()); }
int PhysicalObj::GetY() const { return Arrondit(GetYf()); }
int PhysicalObj::GetCenterX() const 
{ 
  return GetX() +m_test_left +GetTestWidth()/2;
}
int PhysicalObj::GetCenterY() const 
{ 
  return GetY() +m_test_top +GetTestHeight()/2;
}

double PhysicalObj::GetMass() const { return m_mass; }

//-----------------------------------------------------------------------------

double PhysicalObj::GetStrengthX() const
{ assert (!IsGhost()); return m_strength.x; }
double PhysicalObj::GetStrengthY() const
{ assert (!IsGhost()); return m_strength.y; }
double PhysicalObj::GetStrengthLength() const 
{ assert (!IsGhost()); return hypot(m_strength.x, m_strength.y); }
double PhysicalObj::GetStrengthAngle() const 
{
  assert (!IsGhost());
  return CalculeAngle (
    CL_Point(0,0), 
    CL_Point((int)m_strength.x, (int)m_strength.y));
}

//-----------------------------------------------------------------------------

const CL_Point PhysicalObj::GetCenter() const 
{ return CL_Point(GetCenterX(), GetCenterY()); }

//-----------------------------------------------------------------------------

// Lit les dimensions de l'objet
int PhysicalObj::GetWidth() const 
{ 
  assert (m_width != 0); 
  return m_width; 
}
int PhysicalObj::GetHeight() const 
{ 
  assert (m_height != 0); 
  return m_height; 
}
int PhysicalObj::GetTestWidth() const 
{ 
  return m_width -m_test_left -m_test_right; 
}
int PhysicalObj::GetTestHeight() const 
{ 
  return m_height -m_test_top -m_test_bottom; 
}

//-----------------------------------------------------------------------------

const CL_Rect PhysicalObj::GetRect() const 
{ 
  const int x = GetX();
  const int y = GetY();
  return CL_Rect(x, y, x+m_width, y+m_height);
}

//-----------------------------------------------------------------------------

const CL_Rect PhysicalObj::GetTestRect() const 
{ 
  const int x = GetX();
  const int y = GetY();
  return CL_Rect(
    x +m_test_left,       
    y +m_test_top, 
    x +m_width  -m_test_right, 
    y +m_height -m_test_bottom);  
}

//-----------------------------------------------------------------------------

void PhysicalObj::SetX (double x) { SetXY(x, m_pos.y); }
void PhysicalObj::SetY (double y) { SetXY(m_pos.x, y); }
void PhysicalObj::SetWidth (uint width) { m_width = width; }
void PhysicalObj::SetHeight (uint height) { m_height = height; }
void PhysicalObj::SetSize (uint width, uint height) 
{ 
  assert (width != 0);
  assert (height != 0);
  m_width = width; 
  m_height = height; 
}

void PhysicalObj::SetMass (double mass) { m_mass = mass; }

//-----------------------------------------------------------------------------

void PhysicalObj::SetTestRect (uint left, uint right, uint top, uint bottom)
{
  m_test_left =  left;
  m_test_right = right;
  m_test_top = top;
  m_test_bottom = bottom;
}

//-----------------------------------------------------------------------------

void PhysicalObj::SetXY (double x, double y)
{
  if (IsOutsideWorldXY (static_cast<int>(x), static_cast<int>(y))) 
  {
#ifdef DEBUG_CHG_ETAT
    COUT_DEBUG << "SetXY sort du monde : " << x << "," << y << std::endl;
#endif
    Ghost();
  } else {
    m_pos.x = x;
    m_pos.y = y;
    StartMoving();
  }
}

//-----------------------------------------------------------------------------

void PhysicalObj::NewStrength (double length, double angle)
{
  NewStrengthXY (length*cos(angle), length*sin(angle));
}

//-----------------------------------------------------------------------------

void PhysicalObj::NewStrengthXY (double x, double y)
{
#ifdef DEBUG_MSG_PHYSIQUE
  COUT_PHYSIC << "Nouvelle force : " << x << " ," << y << std::endl;
#endif
  if (EgalZero(x) && EgalZero(y))
  {
#ifdef DEBUG_MSG_PHYSIQUE
    COUT_PHYSIC << "Force nulle. On se barre !" << std::endl;
#endif
    return;
  }
  if (!IsMoving())
    m_last_move = Wormux::temps.Lit();
  if (EgalZero(x)) x = 0;
  if (EgalZero(y)) y = 0;
  m_strength.x = x;
  m_strength.y = y;
  StartMoving();
  PhysicalEngine();
}

//-----------------------------------------------------------------------------

void PhysicalObj::NewSpeedXY (double x, double y)
{
  m_speed.x = x; 
  m_speed.y = y;
  m_last_move = Wormux::temps.Lit();
}

//-----------------------------------------------------------------------------

void PhysicalObj::AddStrength (double length, double angle)
{
#ifdef DEBUG_MSG_PHYSIQUE
  COUT_PHYSIC << "Ajoute une force : lg=" 
                      << length << ", angle=" << angle << ")" << std::endl;
#endif

  if (length < EPS_ZERO)
  {
#ifdef DEBUG_MSG_PHYSIQUE
    COUT_DEBUG << "Force nulle, on se barre !" << std::endl;
#endif
    return;
  }

  m_strength.x += length*cos(angle);
  m_strength.y += length*sin(angle);
  StartMoving();
#ifdef DEBUG_MSG_PHYSIQUE
  COUT_PHYSIC << "Force ajoute -> x=" << m_strength.x 
                      << ", y=" << m_strength.y << std::endl;
#endif
}

//-----------------------------------------------------------------------------

void PhysicalObj::CalculateSpeed(double &x, double &y)
{
  if(!IsMoving())
  {
    x = 0;
    y = 0;
    return;
  }
  x = m_speed.x;
  y = m_speed.y;
}

//-----------------------------------------------------------------------------

double PhysicalObj::CalculateMvtAngle()
{
  if (!IsMoving()) return 0.0;
  return CalculeAngle (CL_Point(0,0),CL_Point((int)m_speed.x,(int)m_speed.y));
}

//-----------------------------------------------------------------------------

// Move to a point with collision test
// Return true if collision occured
bool PhysicalObj::MoveCollisionTest (double nv_x, double nv_y)
{
  double x,y,dx,dy;
  int tmp_x, tmp_y;
  const double step= 1.0;

  // Calcule la longueur et les pas d'incrmentation
  double lg = Distance (GetXf(), GetYf(), nv_x, nv_y);
  assert (lg!=0);
  x = m_pos.x;
  y = m_pos.y;
  dx = nv_x - x;
  dy = nv_y - y;
  dx /= lg;
  dx *= step;
  dy /= lg;
  dy *= step;

  do
  {
    tmp_x = Arrondit(x+dx);
    tmp_y = Arrondit(y+dy);
    
    // Si on sort du monde, ne fait aucun test
    if (IsOutsideWorldXY (tmp_x, tmp_y))
    {
      if (!config.exterieur_monde_vide) {
        tmp_x = BorneLong(tmp_x, 0, monde.GetWidth() -GetWidth() -1);
        tmp_y = BorneLong(tmp_y, 0, monde.GetHeight() -GetHeight() -1);
#ifdef DEBUG_CHG_ETAT
	    COUT_DEBUG << "DeplaceTestCollision touche un bord :" 
                   << tmp_x << "," << tmp_y
                   << std::endl;
#endif
        SetXY (tmp_x, tmp_y);
      } else {
#ifdef DEBUG_CHG_ETAT
        COUT_DEBUG << "DeplaceTestCollision sort du monde (en cours)." << std::endl;
#endif
        Ghost();
      }
      return true;
    }
  
    // Test de collision
    if (CollisionTest(tmp_x, tmp_y)) 
    {
#ifdef DEBUG_CHG_ETAT
      COUT_DEBUG << "DeplaceTestCollision: collision par TestCollision."
		 << std::endl;
#endif
      return true;
    }

    // Deplace l'objet
    x += dx;
    y += dy;
    m_pos.x=x;
    m_pos.y=y;
  
    // Iteration
    lg -= step;    
  } while (0 < lg);
  m_pos.x=nv_x;
  m_pos.y=nv_y;
  return false;
}

//-----------------------------------------------------------------------------

void PhysicalObj::StartMoving()
{
  m_last_move = Wormux::temps.Lit();
#ifdef DEBUG_MSG_PHYSIQUE
//  COUT_DEBUG0 << "Start moving." << std::endl;
#endif
}

//-----------------------------------------------------------------------------

void PhysicalObj::StopMoving()
{
  if ((m_speed.y>0.0) && !FootsInVacuum()) SignalFallEnding();
#ifdef DEBUG_MSG_PHYSIQUE
//  if (IsMoving()) COUT_PHYSIC << "Fin d'un mouvement." << std::endl;
#endif
//  m_moving = false;
  m_strength.x = m_strength.y = 0;
  m_speed.x = m_speed.y = 0;
}

//-----------------------------------------------------------------------------

bool PhysicalObj::IsMoving() const
{
  return ( m_speed.x != 0.0
        || m_speed.y != 0.0
        || m_strength.x != 0.0
        || m_strength.y != 0.0);
}

//-----------------------------------------------------------------------------

// Calcule la position d'un point  un moment donn
void PhysicalObj::CalculateXY (double &x, double &y)
{
  x = m_pos.x;
  y = m_pos.y;

  double delta_t = (Wormux::temps.Lit() - m_last_move) / 1000.0;
  
  double acc_x,acc_y;

  acc_x = m_strength.x + (vent.LitVal() * m_wind_factor * delta_t);
  acc_y = m_strength.y + (config.gravity * delta_t * m_mass / 5.0);

  m_speed.x += acc_x;
  m_speed.y += acc_y;

  x += m_speed.x * delta_t;
  y += m_speed.y * delta_t;

  m_last_move = Wormux::temps.Lit();

  //Rinitialisation de la force
  m_strength.x = 0.0;
  m_strength.y = 0.0;
}

//-----------------------------------------------------------------------------

void PhysicalObj::PhysicalEngine ()
{
  // On veut pas de fantome par ici !
  if (IsGhost()) return; 

  if(!IsMoving() && m_type != objCLASSIC) return;

  if ((m_type == objCLASSIC) && !IsMoving() 
      && !FootsInVacuum() && !IsInWater()) return;

  // Calcule les nouvelles coordonnes
  double nv_x,nv_y;
  CalculateXY(nv_x,nv_y);

  // Aucun mouvement ? On se barre !
  if (nv_x == m_pos.x && nv_y == m_pos.y) return;

  if (IsInWater())
    Drown();

  if(m_type == objUNBREAKABLE || IsInWater())
  {
    if(IsOutsideWorldXY((int)nv_x,(int)nv_y))
      Ghost();
    m_pos.x = nv_x;
    m_pos.y = nv_y;
    return;
  }

  // Dplace l'objet en testant les collisions
  bool collision = MoveCollisionTest (nv_x, nv_y);

  // On veut pas de fantome par ici !
  if (IsGhost()) return;
  
  if (collision)
  {
#ifdef DEBUG_MSG_PHYSIQUE
    COUT_PHYSIC << "Collision durant le dplacement "
			<< '(' << m_pos.x << ',' << m_pos.y 
			<< " -> " << nv_x << ',' << nv_y << ")." << std::endl;
#endif
    if (m_rebounding) {
      Rebound();
    } else {
      StopMoving();
      SignalFallEnding();
    }
  }
}

//-----------------------------------------------------------------------------

void PhysicalObj::Ready()
{
#ifdef DEBUG_CHG_ETAT
  if (m_alive != ALIVE) COUT_DEBUG << "Ready." << std::endl;
#endif
  m_alive = ALIVE;
//  StopMoving();
}

//-----------------------------------------------------------------------------

void PhysicalObj::Die()
{
  assert (m_alive == ALIVE || m_alive == DROWNED);
#ifdef DEBUG_CHG_ETAT
  COUT_DEBUG << "Meurt." << std::endl;
#endif
  m_alive = DEAD;
  SignalDeath();
}

//-----------------------------------------------------------------------------

void PhysicalObj::Ghost ()
{
  if (m_alive == GHOST) return;

  bool was_dead = (m_alive == DEAD);
  m_alive = GHOST;
#ifdef DEBUG_CHG_ETAT
  COUT_DEBUG << "Fantome (etait_mort=" << was_dead << ")" << std::endl;
#endif

  // L'objet devient un fantome
  m_speed.y = 0.0;
  StopMoving();

  SignalGhostState(was_dead);
}

//-----------------------------------------------------------------------------

void PhysicalObj::Drown()
{
  if (m_alive == DROWNED) return;
#ifdef DEBUG_CHG_ETAT
  COUT_DEBUG << "Se noie !!!" << std::endl;
#endif
  m_alive = DROWNED;

  //modifie la vitesse
  double x, y;
  CalculateSpeed(x, y);
  m_speed.x = x/4;
  m_speed.y = y/4;
  SignalDrowning();  
}

//-----------------------------------------------------------------------------

bool PhysicalObj::IsReady() const
{ return !IsMoving() && m_ready;}

bool PhysicalObj::IsDead () const
{ return ((m_alive == GHOST) || (m_alive == DEAD)); }

bool PhysicalObj::IsGhost() const 
{ return (m_alive == GHOST); }

bool PhysicalObj::IsDrowned() const 
{ return (m_alive == DROWNED); }

//-----------------------------------------------------------------------------

bool PhysicalObj::IsOutsideWorldXY (int x, int y) const
{
  x += m_test_left;
  y += m_test_top;
  if (monde.EstHorsMondeXlarg(x, GetTestWidth())) return true;
  if (monde.EstHorsMondeYhaut(y, GetTestHeight()))
  {
    if (m_allow_negative_y)
    {
      if ((Y_OBJET_MIN <= y) && (y+GetTestHeight()-1 < 0)) return false;
    }
    return true;
  }
  return false;
}    

//-----------------------------------------------------------------------------

bool PhysicalObj::IsOutsideWorld (int dx, int dy) const
{ return IsOutsideWorldXY (GetX()+dx, GetY()+dy); }

//-----------------------------------------------------------------------------

bool PhysicalObj::FootsOnFloor(int y) const
{
  // if outside is empty, the object can't hit the ground !
  if (config.exterieur_monde_vide) return false;

  const int y_max = monde.GetHeight()-m_height +m_test_bottom;
  return (y_max <= y);
}

//-----------------------------------------------------------------------------

bool PhysicalObj::IsInVacuum (int dx, int dy) const
{
  return IsInVacuumXY (GetX()+dx, GetY()+dy);
}

//-----------------------------------------------------------------------------

bool PhysicalObj::IsInVacuumXY (int x, int y) const
{
  if (IsOutsideWorldXY(x,y)) return config.exterieur_monde_vide;
  
  if (FootsOnFloor(y-1)) return false;

  CL_Rect rect(
    x +m_test_left,       
	y +m_test_top, 
	x +m_width  -m_test_right, 
	y +m_height -m_test_bottom/*-1*/);
  if (m_allow_negative_y)
  {
    if ((0<=rect.bottom) && (rect.top<0)) rect.top = 0;
  }
  return monde.RectEstDansVide (rect);
}

//-----------------------------------------------------------------------------

bool PhysicalObj::FootsInVacuum() const
{
  return FootsInVacuumXY(GetX(), GetY());
}

//-----------------------------------------------------------------------------

bool PhysicalObj::FootsInVacuumXY(int x, int y) const
{
  if (IsOutsideWorldXY(x,y)) return config.exterieur_monde_vide;

  if (FootsOnFloor(y)) return false;
  
  int y_test = y +m_height -m_test_bottom/*-1*/;
  CL_Rect rect(
    x +m_test_left,       
    y_test,
	x +m_width  -m_test_right, 
	y_test+1);
  if (m_allow_negative_y)
  {
    if ((0<=rect.bottom) && (rect.top<0)) rect.top = 0;
  }
  return monde.RectEstDansVide (rect);
}

//-----------------------------------------------------------------------------

bool PhysicalObj::IsInWater () const
{
  if (!eau.EstActive()) return false;
  return (int)eau.GetHeighteur() < GetY();
}

//-----------------------------------------------------------------------------

bool PhysicalObj::CollisionTest (int x, int y)
{ 
  return !IsInVacuumXY (x,y);
}

//-----------------------------------------------------------------------------

void PhysicalObj::DirectFall()
{
  while (!IsGhost() && !IsDrowned() && FootsInVacuum()) SetY(GetY()+1);
}

//-----------------------------------------------------------------------------

void PhysicalObj::Rebound()
{
  assert (m_rebounding);

  //Calcule de la tangeante au point de collision:
  //On cherche le point de contact entre le sol et l'obj;
  bool contact_ok;
  int contact_x,contact_y;  
  double angle_terrain;

#ifdef DEBUG_MSG_PHYSIQUE
  COUT_DEBUG << "Rebondit !" << std::endl
             << "A l'arrive: " << m_speed.x
             << ";" << m_speed.y
             << " / angle: "<< CalculateMvtAngle()
             << std::endl;
#endif

  contact_ok = ContactPoint(contact_x,contact_y);

  if(contact_ok)
    angle_terrain=monde.terrain.Tangeante(contact_x,contact_y);

#ifdef DEBUG_MSG_PHYSIQUE
  COUT_DEBUG << "Angle du terrain : " << angle_terrain
             << std::endl;
#endif
  
  if(angle_terrain == -1.0 || !contact_ok)
  {
    m_speed.x = - m_speed.x * m_rebound_factor;
    m_speed.y = - m_speed.y * m_rebound_factor;
#ifdef DEBUG_MSG_PHYSIQUE
  COUT_DEBUG << "Collision en un point!" << std::endl;
#endif
  }
  else
  {
    double nx,ny;
    double angle = CalculateMvtAngle();
    
    while(angle < -M_PI)
      angle += 2.0 * M_PI;
    while(angle > M_PI)
      angle -= 2.0 * M_PI;
    
    double norme = Norme(m_speed.x,m_speed.y);
    
    //Symtrique de l'angle par rapport   la tangeante
    angle -= 2.0 * (angle - angle_terrain);
    
    nx = norme * cos(angle);
    ny = norme * sin(angle);
    m_speed.x = nx * m_rebound_factor;
    m_speed.y = ny * m_rebound_factor;
  }
#ifdef DEBUG_MSG_PHYSIQUE
  COUT_DEBUG << "En repartant: " << m_speed.x
             << ";" << m_speed.y
             << " / angle: "<< CalculateMvtAngle()
             << std::endl;
#endif
  //Seuil  partir duquel on arrete les rebonds
  if (fabs(m_speed.x)<10.0 &&  fabs(m_speed.y)<10.0)
  {
    StopMoving();
    return;
  }
  
  if (!m_rebound_sound.empty()) jukebox.Play(m_rebound_sound) ;
}

//-----------------------------------------------------------------------------

bool PhysicalObj::ContactPoint (int & contact_x, int & contact_y)
{
  //On cherche un point de contact en bas de l'objet:
  for (uint x=GetX()+ m_test_left; x<=(GetX()+m_width)-m_test_right; x++)
  {
    if(!monde.EstHorsMonde(CL_Point(x,(GetY()+m_height-m_test_bottom))))
    if(!monde.terrain.EstDansVide(x,(GetY()+m_height-m_test_bottom)+1)
    &&( monde.terrain.EstDansVide(x,(GetY()+m_height-m_test_bottom))
    ||  monde.terrain.EstDansVide(x,(GetY()+m_height-m_test_bottom)+2)))
    {
      contact_x = x;
      contact_y = GetY() +m_height-m_test_bottom +1;
      return true;
    }
  }
  //On cherche un point de contact  gauche de l'objet:
  for(uint y=GetY()+m_test_top;y<=GetY()+m_height-m_test_bottom;y++)
  {
    if(!monde.EstHorsMonde(CL_Point(GetX()+m_test_left-1,y)))
    if(!monde.terrain.EstDansVide(GetX()+m_test_left-1,y)
    &&( monde.terrain.EstDansVide(GetX()+m_test_left,y)
    ||  monde.terrain.EstDansVide(GetX()+m_test_left-2,y)))
    {
      contact_x = GetX() +m_test_left-1;
      contact_y = y;
      return true;
    }
  }
  //On cherche un point de contact  droite de l'objet:
  for(uint y=GetY()+m_test_top;y<=GetY()+m_height-m_test_bottom;y++)
  {
    if(!monde.EstHorsMonde(CL_Point((GetX()+m_width-m_test_right)+1,y)))
    if(!monde.terrain.EstDansVide((GetX()+m_width-m_test_right)+1,y)
    &&( monde.terrain.EstDansVide((GetX()+m_width-m_test_right)+2,y)
    ||  monde.terrain.EstDansVide((GetX()+m_width-m_test_right),y)))
    {
      contact_x = GetX() + m_width - m_test_right +1;
      contact_y = y;
      return true;
    }
  }
  //On cherche un point de contact en haut de l'objet:
  for(uint x=GetX()+m_test_left;x<=GetX()+m_width-m_test_right;x++)
  {
    if(!monde.EstHorsMonde(CL_Point(x,GetY()+m_test_top-1)))
    if(!monde.terrain.EstDansVide(x,GetY()+m_test_top-1)
    &&( monde.terrain.EstDansVide(x,GetY()+m_test_top-2)
    ||  monde.terrain.EstDansVide(x,GetY()+m_test_top)))
    {
      contact_x =x;
      contact_y = GetY() +m_test_top;
      return true;
    }
  }
  return false;
}

//-----------------------------------------------------------------------------

