/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Abstract class used for physical object (object with a size, mass,
 * etc.). This object can have differents state : ready, is moving, or ghost
 * (is outside of the world).
 *
 * You can : make the object move (with collision test), change state, etc.
 * If the object go outside of the world, it become a ghost.
 *****************************************************************************/

#ifndef PHYSICAL_OBJECT_H
#define PHYSICAL_OBJECT_H
//-----------------------------------------------------------------------------
#include "../include/base.h"
#include <ClanLib/core.h>  // Pour CL_Point et CL_Rect
//-----------------------------------------------------------------------------

// Alive state
typedef enum
{
  ALIVE,
  DEAD,
  GHOST,
  DROWNED	  
} vivant_t;

//-----------------------------------------------------------------------------

// Type d'objet
typedef enum
{
  // Unbreakable object : detected by collision test, but isn't touch by
  // explosion
  objUNBREAKABLE,

  // Objet classique : dtect pour le test de collision, subit le souffle
  // des explosions, etc.
  objCLASSIC
} type_objet_t;

//-----------------------------------------------------------------------------

class PhysicalObj
{
public:
  // Object name (useful for debug ;-))
  std::string m_name;

  type_objet_t m_type;

  bool m_rebounding;

private:
  // Object size
  uint m_width, m_height;

  // Rectangle used for collision tests
  uint m_test_left, m_test_right, m_test_top, m_test_bottom;

  double m_mass;

protected:
  typedef struct { double x,y; } DoubleVector;
  DoubleVector m_pos;
  DoubleVector m_strength, m_speed;
  uint m_last_move;         // Time since last mouvement

  // Variable utilise par les fils de cet classe
  // permettant de modifier l'tat prt/occup
  // (inutilise par PhysicalObj)
  bool m_ready;

  // Est affect par le vent ?  0 = vent dsactive
  double m_wind_factor;
  
  // L'objet rebondit lors des collisions
  double m_rebound_factor;
  std::string m_rebound_sound;

  vivant_t m_alive;
  bool m_allow_negative_y;

public:
  // Initialise les variables
  PhysicalObj (const std::string &name, double masse=0.0);
  virtual ~PhysicalObj ();

  // Refresh datas
  virtual void Refresh() = 0;

  // Draw the object
  virtual void Draw() = 0;

  // Initialisation of datas (only called once)
  virtual void Init() = 0;

  // Reset datas
  virtual void Reset() = 0;

  //----------- Physic -------------

  void NewStrength (double norme, double angle);
  void NewStrengthXY (double x, double y);
  void NewSpeedXY (double x, double y);

  // Calculate actual speed
  void CalculateSpeed (double &x, double &y);

  // Calculate mouvement angle
  double CalculateMvtAngle ();

  // Add new strength
  void AddStrength (double length, double angle);

  // Physical engine : update position (and state) with current time
  void PhysicalEngine();

  //-------- state ----

  void Ready();
  void Die();
  void Ghost();
  void Drown();
  
  bool IsReady() const;
  bool IsDead() const;
  bool IsGhost() const;
  bool IsDrowned() const;

  //-------- Set position and size -------

  // Move object : if it goes outside the world, it become a ghost
  void SetX (double x);
  void SetY (double y);
  void SetXY (double x, double y);
  void SetWidth (uint width);
  void SetHeight (uint height);
  void SetSize (uint width, uint height);
  void SetMass (double mass);
  void SetTestRect (uint left, uint right, uint top, uint bottom);

  //----------- Access to datas (read only) ----------

  int GetX() const;
  int GetY() const;
  const CL_Point GetPos() const;
  double GetXf() const;
  double GetYf() const;
  int GetCenterX() const;
  int GetCenterY() const;
  const CL_Point GetCenter() const;
  int GetWidth() const;
  int GetHeight() const;
  const CL_Rect GetRect() const;
  double GetMass() const;

  double GetStrengthX() const;
  double GetStrengthY() const;
  double GetStrengthLength() const;
  double GetStrengthAngle() const;

  const CL_Rect GetTestRect() const;
  int GetTestWidth() const;
  int GetTestHeight() const;

  type_objet_t GetObjectType() const { return m_type; }

  bool IsInVacuumXY (int x, int y) const;
  bool IsInVacuum (int dx, int dy) const; // relative to current position
  bool FootsInVacuum () const;
  bool FootsInVacuumXY (int x, int y) const;
  
  bool FootsOnFloor(int y) const;

  bool IsInWater () const;

  // The object is outside of the world
  bool IsOutsideWorldXY (int x, int y) const;
  bool IsOutsideWorld (int dx, int dy) const;

  // The object is moving ?
  bool IsMoving() const;

  // Move to a point with collision test
  // Return true if collision occured
  bool MoveCollisionTest (double nv_x, double nv_y);

protected:
  virtual void SignalDeath();
  virtual void SignalGhostState (bool was_already_dead);
  virtual void SignalFallEnding();
  virtual void SignalDrowning();

  // The object fall directly to the ground (or become a ghost)
  void DirectFall ();

  // Compute current (x,y) position
  virtual void CalculateXY (double &x, double &y);

  // Collision test for point (x,y)
  bool CollisionTest (int x, int y);

private:
  // Stop moving : signal it if the object was falling
  void StopMoving ();

  // Start moving
  void StartMoving();

  //Renvoie la position du point de contact entre
  //l'obj et le terrain
  bool ContactPoint (int &x, int &y);

  // Make the object rebound
  void Rebound();
};

//-----------------------------------------------------------------------------
#endif
