/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Une quipe de vers.
 *****************************************************************************/

#include "team.h"
//-----------------------------------------------------------------------------
#include "../game/game.h"
#include "../game/game_mode.h"
#include "../interface/cursor.h"
#include "../include/constant.h"
#include "../map/camera.h"
#include "../weapon/weapons_list.h"
#include "../weapon/dynamite.h"
#include "../tool/i18n.h"
#include "../tool/file_tools.h"
#include <sstream>
using namespace Wormux;
using namespace std;
//-----------------------------------------------------------------------------

const char* NOM_DEFAUT_EQUIPE = "Team X";

//-----------------------------------------------------------------------------

Team::Team()
{
  vers_fin = vers.end();
}

//-----------------------------------------------------------------------------

bool Team::Init (int nbr_ver, 
		   const std::string &name,
		   const std::string &directory)
{
  std::string nomfich;
  try
  {
    LitDocXml doc;
    m_nomfich = name;

    // Charge le XML
    nomfich = directory + "team.xml";
    if (!FichierExiste(nomfich)) return false;
    if (!doc.Charge (nomfich)) return false;

    CL_ResourceManager res(nomfich, false);

    // Charge les donnes
    cout << "o " << Format(_("Load %s team."), m_nomfich.c_str()) << endl;
    if (!ChargeDonnee (nbr_ver, doc.racine(), &res)) return false;
  }
  catch (const xmlpp::exception &e)
  {
    std::cout << Format(_("Error loading %s team :"),
			name.c_str()) << std::endl
	      << e.what() << std::endl;
    return false;
  }
  return true;
}

//-----------------------------------------------------------------------------

void Team::InitEnergy (uint max)
{
  energie.Init();
  energie.ChoisitNom(m_nom);
  energie.FixeMax(max);
  energie.FixeValeur(LitEnergie());
}

//-----------------------------------------------------------------------------

uint Team::LitEnergie ()
{
  uint total_energie = 0;
  for (const_iterator it=vers.begin(); it != vers_fin; ++it) {
    if( !it->IsDead() )
      total_energie += it->GetEnergy();
  }
  return total_energie;
}
//-----------------------------------------------------------------------------

void Team::ActualiseBarreEnergie ()
{
  energie.NouvelleValeur(LitEnergie());
}

//-----------------------------------------------------------------------------

bool Team::ChargeDonnee (int nbr_ver, xmlpp::Element *xml, 
			   CL_ResourceManager *res)
{
  xml = LitDocXml::AccesBalise (xml, "equipe");
  // Valeurs par dfaut
  camera_est_sauve = false;
  arme_active = &dynamite;

  m_nom = "Team unamed";
  m_profile_son="default";
  viseur.Init();

  // Lit le nom
  if (!LitDocXml::LitString(xml, "nom", m_nom)) return false;

  // Ecusson
  CL_Surface ecusson_tmp("flag", res);
  ecusson = ecusson_tmp;

  // Recupre le nom du profile sonore
  LitDocXml::LitString(xml, "profile_son", m_profile_son);

  // Crer les vers
  xmlpp::Node::NodeList nodes = xml -> get_children("ver");
  xmlpp::Node::NodeList::iterator 
    it=nodes.begin(),
    fin=nodes.end();

  vers.clear();
  bool fin_bcl;
  do
  {
    xmlpp::Element *elem = dynamic_cast<xmlpp::Element*> (*it);
    Skin *skin;
    std::string character_name="Soldat Inconnu";
    std::string skin_name="ver_jaune";
    LitDocXml::LitAttrString(elem, "nom", character_name);
    LitDocXml::LitAttrString(elem, "avatar", skin_name);

    if (skins_list.find(skin_name) != skins_list.end()) {
      skin = &skins_list[skin_name];
    } else {
      cout << Format(_("Error: can't find the skin \"%s\" for the team \"%s\"."),
		     skin_name.c_str(),
		     m_nom.c_str()) << endl;
      return false;
    }

    // Initialise les variables du ver, puis l'ajoute  la liste
    Character new_character;
    vers.push_back(new_character);
    vers.back().InitTeam (this, character_name, skin);

    // C'est la fin ?
    ++it;
    fin_bcl = (it == fin);
    if (nbr_ver != -1) fin_bcl |= (nbr_ver <= (int)vers.size());
  } while (!fin_bcl);

  ver_actif = vers.begin();
  vers_fin = vers.end();
  return (1 <= vers.size());
}

//-----------------------------------------------------------------------------

void Team::internal_NextCharacter()
{
  // Passe au ver suivant
  assert (0 < NbAliveCharacter());
  do
  { 
    ++ver_actif;
    if (ver_actif == vers_fin) {
      ver_actif = vers.begin();
    }
  } while (ActiveCharacter().IsDead());
}

//-----------------------------------------------------------------------------

void Team::NextCharacter()
{
  internal_NextCharacter();

  // Prpare le tour pour le nouveau ver
  camera.ChangeObjSuivi (&ActiveCharacter(), true, true);
  curseur_ver.SuitVerActif();
}

//-----------------------------------------------------------------------------

uint Team::NbAliveCharacter() const
{
  uint nbr = 0;
  for (const_iterator it=vers.begin(); it != vers_fin; ++it)
    if (!it -> IsDead()) ++nbr;
  return nbr;
}

//-----------------------------------------------------------------------------

// Select an *ALIVE* character
void Team::SelectCharacter (Team::iterator it)
{
  // Ver mort ?
  assert (it != vers_fin);
  assert (!it -> IsDead());

  // Change de ver
  ver_actif = it;
  camera.ChangeObjSuivi (&ActiveCharacter(), true, true);
  curseur_ver.SuitVerActif();
}

//-----------------------------------------------------------------------------

// Prepare le tour d'une equipe
void Team::PrepareTour()
{
  // Choisi un ver vivant si possible
  if (ActiveCharacter().IsDead())
  {
    camera_est_sauve = false;
    internal_NextCharacter();
  }

  if (camera_est_sauve) camera.SetXYabs (sauve_camera.x, sauve_camera.y);
  camera.ChangeObjSuivi (&ActiveCharacter(), 
			 !camera_est_sauve, !camera_est_sauve, 
			 true);
  curseur_ver.SuitVerActif();

  // Active notre arme
  (*arme_active).Active();
}

//-----------------------------------------------------------------------------

// Fin d'un tour : nettoyage avant de partir :-)
void Team::FinTour()
{
  // Dsactive notre arme
  (*arme_active).Desactive();
  camera_est_sauve = true;
  sauve_camera = CL_Point(camera.GetX(), camera.GetY());
}

//-----------------------------------------------------------------------------

Character& Team::ActiveCharacter()
{ 
  assert (ver_actif != vers.end());
  assert (ver_actif != vers_fin);
  return *ver_actif;
}

//-----------------------------------------------------------------------------

// Change d'arme
void Team::ChangeArme (Arme * nv_arme)
{
  (*arme_active).Desactive();
  arme_active = nv_arme;

  AccesArme().Active();
}

//-----------------------------------------------------------------------------

int Team::ReadNbBullets() const
{
  // Read in the Map
  // The same method as in AccesNbBullets can't be use on const team !
  std::map<std::string, int>::const_iterator it 
    = m_nb_bullets.find( arme_active->GetName() ) ;
  
  if (it !=  m_nb_bullets.end())  return ( it->second ) ;  
  return 0 ;
}

//-----------------------------------------------------------------------------

int Team::ReadNbBullets(const std::string &weapon_name) const
{
  // Read in the Map
  // The same method as in AccesNbBullets can't be use on const team !
  std::map<std::string, int>::const_iterator it = 
    m_nb_bullets.find( weapon_name );
  
  if (it !=  m_nb_bullets.end()) return ( it->second ) ;  
  return 0 ;
  
}
//-----------------------------------------------------------------------------

int& Team::AccessNbBullets()
{
  // if value not initialized, it initialize to 0 and then return 0
  return m_nb_bullets[ arme_active->GetName() ] ;
}

//-----------------------------------------------------------------------------

Team::iterator Team::Debut() { return vers.begin(); }
Team::iterator Team::Fin() { return vers_fin; }

//-----------------------------------------------------------------------------

void Team::Reset()
{
  // Reset des munitions
  m_nb_bullets.clear();
  std::list<Arme *> l_weapons_list = gst_armes.GetList() ;
  std::list<Arme *>::iterator itw = l_weapons_list.begin(), 
    end = l_weapons_list.end();

  for (; itw != end ; ++itw) {
    m_nb_bullets[ (*itw)->GetName() ] = (*itw)->nb_bullets;
  }
  
  arme_active = &dynamite;
  camera_est_sauve = false;

  // Reset des vers
  vers_fin = vers.end();
  iterator it=vers.begin(), fin=vers.end();
  uint n = 1;
  ver_actif = vers.begin();
  for (; it != fin; ++it, ++n) 
  {
    vers_fin = it;
    ++vers_fin;
    it -> Reset();
    if (n == mode_jeu.nbr_ver_max) break;
  }
  viseur.Reset();
}

//-----------------------------------------------------------------------------

void Team::Draw()
{
  energie.Draw ();
}

//-----------------------------------------------------------------------------

void Team::Update()
{
  energie.Refresh();
}

//-----------------------------------------------------------------------------
