/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Arme dynamite : lorqu'on "tire", un baton de dynamite est lch. Celui
 * explos aprs un laps de temps. La baton fait alors un gros trou dans la
 * carte, souffle les vers qui sont autour en leur faisant perdre de l'nergie.
 *****************************************************************************/

#include "dynamite.h"
//-----------------------------------------------------------------------------
#include "../game/game_mode.h"
#include "../team/teams_list.h"
#include "../tool/i18n.h"
#include "../graphic/graphism.h"
#include "../game/config.h"
#include "../game/game_loop.h"
#include "../sound/jukebox.h"
#include "../weapon/weapon_tools.h"
#include "../object/objects_list.h"
#include "../weapon/weapons_list.h"
//-----------------------------------------------------------------------------
namespace Wormux {
Dynamite dynamite;
//-----------------------------------------------------------------------------

#ifdef DEBUG
  //  #define DEBUG_CADRE_TEST
#endif

//-----------------------------------------------------------------------------

BatonDynamite::BatonDynamite() : ProjectileArme("baton de dynamite")
{}

//-----------------------------------------------------------------------------

void BatonDynamite::Init()
{
  SetMass (dynamite.cfg().masse);

  CL_ResourceManager res(config.data_dir+"weapon/dynamite/config.xml", false);

  image = CL_Sprite("animation", &res);

  double delay = dynamite.cfg().duree/image.get_frame_count()/ 1000.0 ;
  for (int i=0 ; i < image.get_frame_count(); i++)
    image.set_frame_delay(i, delay) ;

  image.set_play_loop(false);

  SetSize (image.get_width(), image.get_height());
  SetTestRect (0, 0, 2, 3);

  explosion.Init(new CL_Sprite("explosion", &res),
		 false, animUNE_BOUCLE, 60);
}

//-----------------------------------------------------------------------------

void BatonDynamite::Reset()
{
  Ready();
  is_active = false;
  explosion_active = false;
  image.restart();
  image.set_frame (0);
  explosion.Reset();
}

//-----------------------------------------------------------------------------

void BatonDynamite::Refresh()
{
  if (!is_active) return;
  bool fin;
  assert (!IsGhost());
  if (!explosion_active) {
    image.update(); 
    fin = image.is_finished();
    if (fin) explosion_active = true;
  } else {
    explosion.Refresh();
    fin = explosion.EstFinie();
    if (fin) is_active = false;
  }
}

//-----------------------------------------------------------------------------

void BatonDynamite::Draw()
{
  if (!is_active) return;
  assert (!IsGhost());

  int x = GetX();
  int y = GetY();
  if (!explosion_active)
    image.draw (x,y); 
  else {
    x -= explosion.GetWidth()/2;
    y -= explosion.GetHeight()/2;
    explosion.Draw (x,y);
  }

#if defined(DEBUG_CADRE_TEST)
  CL_Display::draw_rect (LitRectTest(), CL_Color::red);
#endif
}

//-----------------------------------------------------------------------------

void BatonDynamite::SignalCollision() {}

void BatonDynamite::SignalGhostState (bool) { is_active = false; }

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

Dynamite::Dynamite() : Arme("dynamite")
{
  extra_params = new DynamiteConfig();
}

//-----------------------------------------------------------------------------

void Dynamite::Init()
{
  m_name = _("Dynamite");

  baton.Init();

  CL_ResourceManager res(config.data_dir+"weapon/dynamite/config.xml", false);
  m_image = CL_Surface("weapon",&res);

  impact = CL_Surface("hole", &res);
  icone = CL_Surface("icon", &res);
}

//-----------------------------------------------------------------------------

void Dynamite::Reset()
{
  baton.Reset();
}

//-----------------------------------------------------------------------------

// Pose une dynamite
bool Dynamite::Tire ()
{
  double vx,vy;
  assert (!m_is_active);

  if (!UtiliseMunition()) return false;

  // Ajoute la reprsentation
  int x,y;
  PosArmeXY (x,y);
  baton.Reset ();
  baton.PrepareTir();
  baton.SetXY (x, y);
  lst_objets.AjouteObjet (&baton, true);

  // Ajoute la vitesse actuelle du ver
  ActiveCharacter().CalculateSpeed (vx, vy);
  baton.NewSpeedXY (vx, vy);

  // Active l'animation
  m_is_active = true;
  jukebox.Play("weapon/dynamite_fuze");
  return true;
}

//-----------------------------------------------------------------------------

void Dynamite::Refresh()
{
  if (m_is_active) {
    // Fin de l'explwosion ?
    if (!baton.is_active) FinExplosion ();
  } else {
    // Change le sens de l'image si ncessaire
    m_image.set_scale(ActiveCharacter().GetDirection(), 1);
  }
}

//-----------------------------------------------------------------------------

void Dynamite::FinExplosion ()
{
  m_is_active = false;

  lst_objets.RetireObjet (&baton);

  jukebox.Stop("weapon/dynamite_fuze");
  
  // Si la dynamite est sortie de l'cran, on ne fait rien
  if (baton.IsGhost()) return;

  // Applique les degats aux vers
  CL_Point centre = baton.GetCenter();
  centre.y = baton.GetY()+baton.GetHeight();
  AppliqueExplosion (centre, centre, impact, cfg(), NULL, "weapon/dynamite_exp");
}

//-----------------------------------------------------------------------------

#ifdef WEAPON_CLEAN
void Dynamite::PosArmeXY (uint &x, uint &y) const
{
  const int dx = ActiveCharacter().LitAvatar().hand.dx;
  const int dy = ActiveCharacter().LitAvatar().hand.dy;
  x = ActiveCharacter().GetX();
  y = ActiveCharacter().GetY()+dy;
  if (ActiveCharacter().GetDirection() == 1)
    x += dx;
  else
    x += ActiveCharacter().GetWidth()-dx-image.get_width();
}

//-----------------------------------------------------------------------------


void Dynamite::Draw()
{
  if (m_is_active || (boucle_jeu.LitEtat() == jeuVER_A_JOUE)) return;

  uint x,y;
  PosArmeXY (x,y);
  image.draw(x,y);
}
#endif

//-----------------------------------------------------------------------------

DynamiteConfig& Dynamite::cfg() 
{ return static_cast<DynamiteConfig&>(*extra_params); }

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

DynamiteConfig::DynamiteConfig()
{
  duree = 2000;
}

void DynamiteConfig::LoadXml(xmlpp::Element *elem)
{
  ExplosiveWeaponConfig::LoadXml(elem);
  LitDocXml::LitUint (elem, "duree", duree);
}

//-----------------------------------------------------------------------------
} // namespace Wormux
