/******************************************************************************
 *  Wormux, a free clone of the game Worms from Team17.
 *  Copyright (C) 2001-2004 Lawrence Azzoug.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************
 * Classes virtuelles permettant de dfinir une arme et un projectile. Les
 * armes ont un nom, une image, un tat actif/inactif et une icne (affiche
 * dans l'interface). Les projectiles sont des objets physiques qui ont un
 * comportement spcial lorsqu'ils entrent en collision ou qu'ils sortent du
 * terrain.
 *****************************************************************************/

#ifndef WEAPON_H
#define WEAPON_H
//-----------------------------------------------------------------------------
#include "../include/base.h"
#include "../object/physical_obj.h"
#include "../graphic/animation.h"
#include "../graphic/image.h"
#include "../team/character.h"
#include "weapon_cfg.h"
#include <string>
#include <ClanLib/display.h>
//-----------------------------------------------------------------------------

// Constante munitions illimites
extern const int MUNITION_ILLIMITEES;

//-----------------------------------------------------------------------------

// Projectil d'une arme
class ProjectileArme : public PhysicalObj
{
public:
  bool is_active;
  CL_Sprite image;

protected:
  Character* dernier_ver_touche;
  PhysicalObj* dernier_obj_touche;

  // Peut toucher les vers et les objets ? (test de collision)
  bool touche_ver_objet;

public:
  ProjectileArme (const std::string &nom);
  virtual void Draw();
  virtual void Refresh();
  virtual void Reset() = 0;
  virtual void Init() = 0;
  void PrepareTir();
  virtual bool CollisionTest (int dx, int dy);
  Character* LitDernierVerTouche() const { return dernier_ver_touche; }
  PhysicalObj* LitDernierObjTouche() const { return dernier_obj_touche; }

  // Il y a eu impact avec un ver, un objet ou le sol ?
  bool TestImpact ();

protected:
  virtual void SignalGhostState (bool was_dead);
  virtual void SignalFallEnding();
  virtual void SignalCollision() = 0;
};

//-----------------------------------------------------------------------------

class Arme
{
protected:
  typedef enum {
    weapon_origin_HAND,
    weapon_origin_OVER
  } weapon_origin_t;
  std::string m_id;
  std::string m_name;
  bool m_is_active;
  CL_Surface m_image;
  struct {
    int dx, dy;
    weapon_origin_t origin;
  } position;

  // Actual strength of the weapon
  uint m_strength;

  // time of beginning to load (for choosing the strength)
  uint m_first_time_loading;

  // how many times can we use the weapon _AGAIN_ in the same turn
  // if m_can_change_weapon == true, this MUST BE >= 1 !
  uint m_nb_replay;

  // how many times the weapon has been used in THIS turn
  uint m_nb_playing;

  // change weapon after ? (for the ninja cord = true)
  bool m_can_change_weapon;

  // Extra parameters
  EmptyWeaponConfig *extra_params;

public:
  // Icone de l'arme dans l'interface
  CL_Surface icone;

  // how many times can we use this weapon ?
  int nb_bullets;

  // Weapon's category (used for key-bindings)
  uint category;

  // if max_strength != 0, display the strength bar
  uint max_strength;

  // Draw the weapon ?
  bool m_draw_weapon;

  // Angle in degrees between -90 to 90
  int min_angle, max_angle;

protected:
  virtual void Refresh() = 0;
  virtual bool Tire() = 0;
  
public:
  Arme (const std::string &id);
  virtual ~Arme() {}

  virtual void Reset() = 0;
  virtual void Init() = 0;

  // Active/dsactive l'arme
  virtual void Active();
  virtual void Desactive();

  // Gestion de l'arme
  void Manage();
  bool CanChangeWeapon() const ;

  // Affichage de l'arme
  virtual void Draw();

  // Utilise une munition
  // Renvoie false si on ne peut pas tirer (munition insuffisante)
  bool MunitionSuffisante() const;
  bool UtiliseMunition();

  // Calcule la position de l'image de l'arme,
  // et de la sortie du canon
  virtual void PosArmeXY (int &x, int &y) const;

  // Tire avec l'arme
  // Renvoie true si l'arme a pu tre enclanche
  bool Shoot();

  // L'arme est encore active (animation par ex.) ?
  bool IsActive() const;

  // update strength (so the strength bar can be updated)
  void UpdateStrength();
  
  // the weapon is ready to use ? (is there bullets left ?)
  virtual bool IsReady() const ;

  // Begin to load, to choose the strength
  virtual void InitLoading() ;

  // Are we loading to choose the strength
  virtual bool IsLoading() const;

  // Stop loading
  virtual void StopLoading() ;

  // Get the rotation point of the weapon
  // This is used for drawing the crosshair
  virtual void RotationPointXY (int &x, int &y) const;

  // Choose a target.
  // Return false if it not fire directly after
  virtual bool ChooseTarget () {return false ;}

  // Load parameters from the xml config file
  // Return true if xml has been succesfully load
  bool LoadXml(xmlpp::Element * weapon);  

  // return the strength of the weapon
  const uint ReadStrength() const;

  // Accs aux donnes
  const std::string& GetName() const;
};

//-----------------------------------------------------------------------------
#endif
